#ifdef USE_DNOTIFY
#define _GNU_SOURCE
#endif
#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <libgen.h>
#include <sys/mman.h>
#include <limits.h>
#include <stdarg.h>
#include <getopt.h>
#include <ctype.h>
#ifdef USE_FAM
#include <fam.h>
#endif

#include "turbotail.h"

#define min(x, y)	((x) < (y) ? (x) : (y))

file *files = NULL;
int n_files = 0;
double sleep_interval = SLEEP_INTERVAL;	/* sleep interval in s */
char quiet = QM_OTHERFILE;
int initial_data = INITIAL_N_LINES;
char initial_mode = IM_LINES;
char initial_direction = ID_END;
char forever = 0;
#ifdef USE_FAM
FAMConnection famc;
#endif

int find_path_max(void)
{
#ifdef PATH_MAX
	int path_max = PATH_MAX;
#else
	int path_max = pathconf("/", _PC_PATH_MAX);
	if (path_max <= 0)
	{
		if (errno) error_exit("find_path_max: pathconf(_PC_PATH_MAX) failed\n");

		path_max = 4096;
	}
	else
		path_max++; /* since its relative to root */
#endif
	if (path_max > 4096)
		path_max = 4096;

	return path_max;
}

void error(char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	(void)vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "errno: %d=%s\n", errno, strerror(errno));

	exit(EXIT_FAILURE);
}

int my_write(int fd, char *buffer, off64_t size)
{
	while(size > 0)
	{
		int rc = write(fd, buffer, min(SSIZE_MAX, size));

		if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			error("write failed\n");

			return -1;
		}
		else if (rc == 0)
		{
			return -1;
		}

		buffer += rc;
		size -= rc;
	}

	return 0;
}

void print_header(file *f)
{
	fflush(stdout);
	printf("\n==> %s <==\n", f -> filename);
	fflush(stdout);
}

int seek_lines_from_start(int fd, int how_many, off64_t *new_offset)
{
	off64_t offset = 0;
	int index = 0;

	if (how_many > 0)
	{
		how_many--;	/* don't ask. oh, ok: I was expecting tail -n +1 to start from line 1,
				 * skipping the first line (and thus showing the 2nd line in the file)
				 * but it is not. so that's why I'm decreasing with one
				 */
	}

	while(how_many > 0)
	{
		char buffer[BUFFER_SIZE];

		/* read ... */
		int rc = read(fd, buffer, BUFFER_SIZE);
		if (rc == 0)
			break;
		else if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			error("read failed\n");

			return -1;
		}

		while(how_many > 0 && index < rc)
		{
			if (buffer[index++] == '\n')
				how_many--;
		}

		if (how_many)
			offset += (off64_t)512;
	}

	/* found them? then we're somewhere in the file but
	 * probably at the wrong offset (due to the 'BUFFER_SIZE'
	 * reads)
	 */
	if (how_many == 0)
	{
		offset += index;
	}

	*new_offset = offset;

	return 0;
}

int seek_lines_from_end(int fd, int how_many, off64_t file_size, off64_t *new_offset)
{
	char buffer[BUFFER_SIZE], *dummy;
	off64_t offset = file_size;
	int index = 0;

	/* seek to the \n before the 'how_many'th line ending with \n */
	while(how_many > -1 && offset > 0)
	{
		int rc;
		int cbs = BUFFER_SIZE, dummy_cbs;

		/* determine current block size */
		if (offset < BUFFER_SIZE)
			cbs = (int)offset;

		/* seek to start of current block */
		offset -= (off64_t)cbs;

		if (lseek64(fd, offset, SEEK_SET) == -1)
		{
			error("lseek64 failed\n");
			return -1;
		}

		/* read a block */
		dummy_cbs = cbs;
		dummy = buffer;
		do
		{
			rc = read(fd, dummy, dummy_cbs);
			if (rc == -1)
			{
				if (errno == EINTR || errno == EAGAIN)
					continue;

				error("read failed\n");

				return -1;
			}
			else if (rc == 0)
			{
				break;
			}

			dummy += rc;
			dummy_cbs -= rc;
		}
		while(dummy_cbs > 0);

		/* count the \n's in this block */
		index = cbs;
		while(index > 0)
		{
			index--;

			if (buffer[index] == '\n')
			{
				how_many--;

				if (how_many == -1)
					break;
			}
		}
	}

	/* found the end of the line before the line we're
	 * looking for, we then only have to seek 1 position
	 * forward to get the start of the line we're looking
	 * for
	 */
	if (how_many == -1)
	{
		offset += (off64_t)(index + 1);
	}
	/* otherwhise, the offset should be 0 */

	*new_offset = offset;

	return 0;
}

int initial_pipechar(file *f)
{
	char buffer[BUFFER_SIZE];

	for(;;)
	{
		int rc = read(f -> fd, buffer, BUFFER_SIZE);
		if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			error("read failed\n");
			return -1;
		}
		else if (rc == 0)
		{
			break;
		}

		if (my_write(1, buffer, rc) == -1)
		{
			return -1;
		}
	}

	return 0;
}

int initial_tail(file *f, char mode, int how_much)
{
	char buffer[BUFFER_SIZE];
	struct stat64 finfo;
	off64_t new_offset = -1;

	if (fstat64(f -> fd, &finfo) == -1)
	{
		error("fstat64 failed\n");
		return -1;
	}

	/* these cannot be seeked in */
	if (S_ISFIFO(finfo.st_mode) || S_ISCHR(finfo.st_mode) || f -> fd == 0)
	{
		return initial_pipechar(f);
	}

	/* seek to start of data */
	if (mode == IM_BYTES)
	{
		if (initial_direction == ID_START)
		{
			new_offset = (off64_t)abs(how_much > 0 ? how_much - 1 : how_much);
		}
		else
		{
			new_offset = (off64_t)(finfo.st_size - how_much);
		}
	}
	else	/* lines */
	{
		int rc;

		if (initial_direction == ID_START)
		{
			rc = seek_lines_from_start(f -> fd, abs(how_much), &new_offset);
		}
		else
		{
			rc = seek_lines_from_end(f -> fd, how_much, finfo.st_size, &new_offset);
		}

		if (rc == -1)
			return -1;
	}

	/* do output */
	if (lseek64(f -> fd, new_offset, SEEK_SET) == -1)
	{
		error("lseek64 failed\n");
		return -1;
	}

	for(;;)
	{
		int rc = read(f -> fd, buffer, BUFFER_SIZE);

		if (rc == 0)	/* EOF */
			break;
		else if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			error("read failed\n");

			return -1;
		}

		if (my_write(1, buffer, rc) == -1)
			return -1;
	}

	return 0;
}

int check_file(file *f)
{
	char different_file;
	char no_new_file_yet = 0;
	struct stat64 new;

	/* get info on the file named 'f -> filename' */
	if (stat64(f -> filename, &new) == -1)
	{
		/* if that file (no longer) exists and we want to retry
		 * opens, close the fd (if open).
		 * otherwhise there's an error situation
		 */
		if (errno == ENOENT && (f -> retry_open || f -> follow_mode == FM_FILENAME))
		{
			/* ... */

			no_new_file_yet = 1;

			if (f -> fd != -1)
			{
				close(f -> fd);
				f -> fd = -1;
			}
		}
		else
		{
			error("stat64 failed\n");
			return -1;
		}
	}

	different_file = (f -> follow_mode == FM_FILENAME || f -> retry_open) && (new.st_dev != f -> d || new.st_ino != f -> i || f -> fd == -1);

	/* see if the file got truncated */
	if (f -> fd != -1 && new.st_size < f -> st_size && quiet != QM_QUIET && different_file == 0)
	{
		if (lseek64(f -> fd, (off64_t)new.st_size, SEEK_SET) == -1)
		{
			error("lseek64 failed\n");
			return -1;
		}

		printf("==> %s: file truncated <==\n", f -> filename);
		fflush(stdout);
	}

	f -> st_size = new.st_size;

	/* device/inode of file is different from previous or there was
	 * no file open?? then close the previous file and open the new
	 * one
	 */
	if (different_file && !no_new_file_yet)
	{
		if (f -> fd != -1)
			close(f -> fd);

		/* we can only tail files */
		if (S_ISREG(new.st_mode))
		{
			f -> fd = open64(f -> filename, O_RDONLY);
			if (f -> fd == -1)
			{
				if (errno == ENOENT)
					return 0;

				return -1;
			}

			/* we're not re-retrieving the stat-structure
			 * after the open: if things got changed between
			 * the stat and the open, the next time this one
			 * is called will see a change again and will
			 * output a message to the user. that is a good
			 * thing
			 */
			f -> d = new.st_dev;
			f -> i = new.st_ino;
			f -> st_size = new.st_size;

			/* ... emit message */
			printf("tail: `%s' has been replaced;  following end of new file\n", f -> filename);
			fflush(stdout);

			return initial_tail(f, initial_mode, initial_data);
		}
		else
		{
			f -> ignore = 1;

			printf("File %s got replaced by an untailable file.\n", f -> filename);
			fflush(stdout);
		}
	}

	return 0;
}

int setup_file(char *filename, file *f, int nr, char follow_mode, char retry_open)
{
	struct stat64 finfo;

#ifdef USE_DNOTIFY
	int signr = SIGRTMIN + (nr % (SIGRTMAX - SIGRTMIN));

	f -> signr = signr;
#endif
#ifdef USE_FAM
	char *real_fname = (char *)malloc(find_path_max() + 1);
	if (!real_fname)
		error("out of memory\n");
	if (!realpath(filename, real_fname))
		error("realpath() failed\n");
	if (FAMMonitorFile(&famc, real_fname, &f -> fr, (void *)f) == -1)
	{
		error("FAMMonitorFile failed: %s (%d)\n", FamErrlist[FAMErrno], FAMErrno);
		return -1;
	}
	free(real_fname);
#endif

	/* get info on file */
	if (stat64(filename, &finfo) == -1)
	{
		if (!(errno == ENOENT && retry_open))
		{
			error("stat on %s failed\n", filename);
			return -1;
		}
	}
	else 
	{
		if (!S_ISREG(finfo.st_mode))
		{
			error("%s is not a regular file\n", filename);
			return -1;
		}

		f -> d = finfo.st_dev;
		f -> i = finfo.st_ino;
		f -> st_size = finfo.st_size;
	}

	f -> filename = strdup(filename);

	f -> follow_mode = follow_mode;
	f -> retry_open  = retry_open;

	f -> fd = open(f -> filename, O_RDONLY);
	if (f -> fd == -1)
	{
		if (!f -> retry_open)
		{
			error("Failed to open %s\n", f -> filename);
			return -1;
		}
	}

#ifdef USE_DNOTIFY
	f -> dir_fd = open(dirname(filename), O_RDONLY);
	if (f -> dir_fd == -1)
	{
		error("Failed to open directory %s\n", dirname(filename));
		return -1;
	}

	if (fcntl(f -> dir_fd, F_SETSIG, SIGRTMIN) == -1)
	{
		error("Failed to do fcntl(F_SETSIG)\n");
		return -1;
	}

	if (fcntl(f -> dir_fd, F_NOTIFY, DN_CREATE|DN_DELETE|DN_RENAME|DN_MODIFY|DN_MULTISHOT) == -1)
	{
		error("Failed to do fcntl(F_NOTIFY)\n");
		return -1;
	}
#endif

	return 0;
}

#ifdef USE_DNOTIFY
static void sig_handler(int sig, siginfo_t *si, void *data)
{
	int loop;

	for(loop=0; loop<n_files; loop++)
	{
		if (
				files[loop].signr == sig &&
				files[loop].ignore == 0 &&
				files[loop].follow_mode == FM_FILENAME)
		{
			if (check_file(&files[loop]) == -1)
			{
				/* ... */
			}
		}
	}
}
#endif

void main_loop(file *files, int n_files)
{
	char buffer[BUFFER_SIZE];
#ifdef USE_FAM
	file *previously_shown = NULL;

	for(;;)
	{
		FAMEvent fe;
		fd_set rfds;

		FD_ZERO(&rfds);
		FD_SET(famc.fd, &rfds);

		if (select(famc.fd + 1, &rfds, NULL, NULL, NULL) == -1)
		{
			if (errno != EINTR && errno != EAGAIN)
			{
				/* ... FIXME ... */
			}

			continue;
		}

		for(;;)
		{
			file *f;
			int rc;
			if (FAMPending(&famc) != 1)
				break;

			if (FAMNextEvent(&famc, &fe) != 1)
				break;

			f = (file *)fe.userdata;

			if (f -> ignore)
				continue;

			if (f -> follow_mode == FM_ONESHOT)
				continue;

			if (f -> fd == -1)
				continue;

			if (check_file(f) == -1)
			{
				/* ... FIXME ... */
			}

			for(;;)
			{
				rc = read(f -> fd, buffer, BUFFER_SIZE);
				if (rc == 0)
				{
					/* continue, file did not change */
					break;
				}
				else if (rc == -1)
				{
					error("Failed reading from file %s\n", f -> filename);
				}
				else
				{
					if ((previously_shown != f && n_files > 1 && quiet == QM_OTHERFILE) || quiet == QM_ALWAYS)
					{
						print_header(f);
					}

					if (my_write(1, buffer, rc) == -1)
						break;

					previously_shown = f;
				}
			}
		}
	}
#else
	int previously_shown = -1;
	int n_not_ignored;

	do
	{
		int loop;

		n_not_ignored = 0;

		for(loop=0; loop<n_files; loop++)
		{
			int rc;

			if (files[loop].ignore)
				continue;

			n_not_ignored++;

			if (files[loop].follow_mode == FM_ONESHOT)
				continue;

			if (files[loop].fd == -1)
				continue;

			if (check_file(&files[loop]) == -1)
			{
				/* ... FIXME ... */
			}

			for(;;)
			{
				rc = read(files[loop].fd, buffer, BUFFER_SIZE);
				if (rc == 0)
				{
					/* continue, file did not change */
					break;
				}
				else if (rc == -1)
				{
					error("Failed reading from file %s\n", files[loop].filename);
				}
				else
				{
					if ((previously_shown != loop && n_files > 1 && quiet == QM_OTHERFILE) || quiet == QM_ALWAYS)
					{
						print_header(&files[loop]);
					}

					if (my_write(1, buffer, rc) == -1)
						break;

					previously_shown = loop;
				}
			}
		}

		if (sleep_interval > 2147.483647)
			sleep(sleep_interval);
		else
			usleep(sleep_interval * 1000000);
	}
	while(n_not_ignored);
#endif
}

int convert_amount(char *str, int *amount, char *from_end)
{
	int multiplier = 1;
	int len = strlen(str);
	int rc = 0;

	if (len == 0)
		return 0;

	*from_end = ID_END;

	if (str[0] == '+')
	{
		*from_end = ID_START;
		str++;
	}
	else if (str[0] == '-')
	{
		str++;
	}

	if (str[len - 1] == 'b')
	{
		multiplier = 512;
		rc = 1;
	}
	else if (str[len - 1] == 'k')
	{
		multiplier = 1024;
		rc = 1;
	}
	else if (str[len - 1] == 'm')
	{
		multiplier = 1024 * 1024;
		rc = 1;
	}

	*amount = multiplier * atoi(str);

	return rc;
}

char parse_obsolete_arguments(int argc, char *argv[], char *follow_mode)
{
	char *par = argv[1];

	if (par[0] == '+' || /* start from the beginning */
			isdigit(par[1])) /* -NN */
	{
		/* start at the start of the file (instead of the end) */
		if (par[0] == '+')
		{
			initial_direction = ID_START;
		}

		/* after the '-' or '+' a number follows */
		initial_data = 0;
		par++; /* start at first digit */
		while(isdigit(*par))
		{
			initial_data *= 10;
			initial_data += (*par) - '0';
			par++;
		}

		/* then follows l or c/b indicating lines or characters */
		if (*par == 'l') /* lines */
		{
			initial_mode = IM_LINES;
			par++;
		}
		else if (*par == 'c' || *par == 'b') /* chars */
		{
			initial_mode = IM_BYTES;
			par++;
		}

		/* 'f' (or nothing) indicating follow */
		if (*par == 'f')
		{
			*follow_mode = FM_DESCRIPTOR;
			forever = 1;
		}

		optind = 2;

		return 1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	int loop;
	char follow_mode = FM_ONESHOT;
	char retry_open = 0;
	int n_files_arg;
	char **arg_files;
	int c;
#ifdef USE_DNOTIFY
	struct sigaction act;

	act.sa_sigaction = sig_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_SIGINFO;
	sigaction(SIGRTMIN, &act, NULL);
#endif

	if (argc > 1 && parse_obsolete_arguments(argc, argv, &follow_mode) == 0)
	{
		while((c = getopt_long(argc, argv, "c:n:fFqs:v", long_options, NULL)) != -1)
		{
			switch(c)
			{
				case 'f':
					if (follow_mode == FM_ONESHOT)
						follow_mode = FM_DESCRIPTOR;
					forever = 1;
					break;

				case LONG_FOLLOW_OPTION:
					if (strcmp(optarg, "descriptor") == 0)
						follow_mode = FM_DESCRIPTOR;
					else if (strcmp(optarg, "name") == 0)
						follow_mode = FM_FILENAME;
					else
					{
						error("%s is an unknown parameter to \"-follow\"\n", optarg);
					}
					forever = 1;
					break;

				case 'F':
					follow_mode = FM_FILENAME;
					break;

				case RETRY_OPTION:
					retry_open = 1;
					quiet = QM_QUIET;
					break;

				case 'q':
					quiet = QM_QUIET;
					break;

				case 'v':
					quiet = QM_ALWAYS;
					break;

				case 's':
					sleep_interval = atof(optarg);
					break;

				case 'n':
					if (convert_amount(optarg, &initial_data, &initial_direction) == 0)
						initial_mode = IM_LINES;
					else
						initial_mode = IM_BYTES;
					break;

				case 'c':
					(void)convert_amount(optarg, &initial_data, &initial_direction);
					initial_mode = IM_BYTES;
					break;

				case 'h':
					printf("Go and read the manpage!\n");
					return 1;

				case MAX_UNCHANGED_STATS_OPTION:
					fprintf(stderr, "--max-unchanged-stats is ignored!\n");
					break;

				case MAX_CONSECUTIVE_SIZE_CHANGES_OPTION:
					fprintf(stderr, "--max-consecutive-size-changes is ignored!\n");
					break;

				case 'V':
					printf("(turbo-)tail by folkert@vanheusden.com, " VERSION "\n");
					return 0;

				default:
					printf("Go and read the manpage!\n");
					return 1;
			}
		}
	}

#ifdef USE_FAM
	if (FAMOpen(&famc) == -1)
	{
		error("FAMOpen failed: %s (%d)\n", FamErrlist[FAMErrno], FAMErrno);
		exit(EXIT_FAILURE);
	}
#endif

	n_files_arg = argc - optind;
	arg_files = argv + optind;
	for(loop=0; loop<n_files_arg; loop++)
	{
		n_files++;

		files = (file *)realloc(files, n_files * sizeof(file));
		memset(&files[n_files - 1], 0x00, sizeof(file));

		if (setup_file(arg_files[loop], &files[n_files - 1], n_files-1, follow_mode, retry_open) == -1)
		{
			exit(EXIT_FAILURE);
		}
	}

	/* no file given? then it must be stdio */
	if (n_files == 0)
	{
		n_files++;

		files = (file *)realloc(files, n_files * sizeof(file));
		memset(&files[n_files - 1], 0x00, sizeof(file));

		files[n_files - 1].fd = 0;
		files[n_files - 1].follow_mode = FM_DESCRIPTOR;
		files[n_files - 1].retry_open = 0;
	}

	/* print the last part of the given files */
	for(loop=0; loop<n_files; loop++)
	{
		if (quiet != QM_QUIET && n_files > 1)
		{
			print_header(&files[loop]);
		}

		if (initial_tail(&files[loop], initial_mode, initial_data) == -1)
		{
			exit(EXIT_FAILURE);
		}
	}

	if (forever)
	{
		main_loop(files, n_files);
	}

	exit(EXIT_SUCCESS);
	return 0;
}
